# Create Flask application
"""Minimal Flask application example.
SPHINX-START
First install examples, setup the application and load
fixture data by running:
.. code-block:: console
   $ pip install -e .[all]
   $ cd examples
   $ ./app-setup.sh
   $ ./app-fixtures.sh  # put your OPENIDC client details here before
Next, start the development server:
.. code-block:: console
   $ export FLASK_APP=app.py FLASK_DEBUG=1
   $ flask run
and open the example application in your browser:
.. code-block:: console
    $ open http://127.0.0.1:5000/
To reset the example application run:
.. code-block:: console
    $ ./app-teardown.sh
SPHINX-END
"""
from __future__ import absolute_import, print_function

import os

from flask_login import current_user
from invenio_accounts import InvenioAccounts
from invenio_db import InvenioDB
from flask_oauthlib.client import OAuth as FlaskOAuth
from invenio_oauthclient import InvenioOAuthClient, InvenioOAuthClientREST
from invenio_oauthclient.views.client import rest_blueprint

from cesnet_openid_remote import CESNETOpenIDRemote
from flask import Flask, redirect, url_for
from flask_babelex import Babel

app = Flask(__name__)

app.config['SQLALCHEMY_DATABASE_URI'] = os.environ.get(
    'SQLALCHEMY_DATABASE_URI', 'sqlite:///app.db'
)

app.config.update(
    ACCOUNTS_USE_CELERY=False,
    CELERY_ALWAYS_EAGER=True,
    CELERY_CACHE_BACKEND='memory',
    CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
    CELERY_RESULT_BACKEND='cache',
    MAIL_SUPPRESS_SEND=True,
    SECRET_KEY='CHANGE_ME',
    SECURITY_PASSWORD_SALT='CHANGE_ME_ALSO',
    SQLALCHEMY_TRACK_MODIFICATIONS=False,
)

Babel(app)
InvenioDB(app)
InvenioAccounts(app)
FlaskOAuth(app)
InvenioOAuthClient(app)
InvenioOAuthClientREST(app)
CESNETOpenIDRemote(app)

app.register_blueprint(rest_blueprint)


@app.route('/cesnet')
def cesnet():
    """Home page: try to print user email or redirect to login with cern."""
    if not current_user.is_authenticated:
        return redirect(url_for('invenio_oauthclient.rest_login',
                                remote_app='eduid'))

    return 'hello {}'.format(current_user)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CESNET OIDC Auth backend for OARepo"""
from urllib.parse import urlparse, parse_qs

from flask import current_app
from invenio_accounts.proxies import current_datastore
from werkzeug.local import LocalProxy

CESNET_OPENID_REMOTE_GROUP_PREFIX = 'urn:geant:cesnet.cz:'
"""Default prefix of group attribute URIs."""

CESNET_OPENID_REMOTE_GROUP_AUTHORITY = 'perun.cesnet.cz'
"""Default authority that issues the group attribute URIs."""

gconf = LocalProxy(
    lambda: dict(
        prefix=current_app.config.get(
            "CESNET_OPENID_REMOTE_GROUP_PREFIX",
            CESNET_OPENID_REMOTE_GROUP_PREFIX),
        authority=current_app.config.get(
            "CESNET_OPENID_REMOTE_GROUP_AUTHORITY",
            CESNET_OPENID_REMOTE_GROUP_AUTHORITY)))


def validate_group_uri(group_uri):
    """Checks if group URI is well-formatted and valid.

       @param group_uri: group URI string
       @returns: True if group URI is valid, False otherwise
    """
    if not group_uri.startswith(gconf['prefix']) or not group_uri.endswith(f'#{gconf["authority"]}'):
        return False
    if 'groupAttributes:' not in group_uri:
        return False

    return True


def parse_group_uri(group_uri):
    """Parses UUID and any extra data from the group URI string.

        @param group_uri: group URI string
        @returns Tuple with (UUID, dict(extra_data)) specification of the group
    """
    attrs = group_uri[len(gconf['prefix']) + len('groupAttributes:'):-(len(gconf['authority']) + 1)]
    parsed_url = urlparse(attrs)
    qs = parse_qs(parsed_url.query)
    if 'displayName' in qs:
        qs['displayName'] = qs['displayName'][0]
    return parsed_url.path, qs



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CESNET OIDC Auth backend for OARepo."""
from datetime import datetime

from flask import current_app, g, session
from flask_login import current_user
from flask_principal import AnonymousIdentity
from invenio_db import db
from invenio_oauthclient.handlers import token_getter
from invenio_oauthclient.handlers.rest import response_handler, default_remote_response_handler
from invenio_oauthclient.models import RemoteAccount, RemoteToken
from invenio_oauthclient.utils import oauth_unlink_external_id, oauth_link_external_id, obj_or_import_string
from invenio_openid_connect import InvenioAuthOpenIdRemote
from werkzeug.local import LocalProxy

from cesnet_openid_remote.constants import OPENIDC_GROUPS_KEY, OPENIDC_CONFIG, OPENIDC_GROUPS_SCOPE
from cesnet_openid_remote.proxies import current_cesnet_openid

"""Pre-configured remote application for enabling sign in/up with eduID federated accounts."""

OAUTHCLIENT_CESNET_HIDDEN_GROUPS = ()
"""Tunable list of groups that will be invisible to this app."""

OAUTHCLIENT_CESNET_HIDDEN_GROUPS_RE = ()
"""Tunable list of regexps of groups to invisible to this app."""


class CesnetOpenIdRemote(InvenioAuthOpenIdRemote):
    """CESNET OIDC Remote Auth backend for OARepo."""
    CONFIG_OPENID = OPENIDC_CONFIG

    name = 'CESNET eduID Login'
    description = 'Log In with your eduID account'

    def __init__(self):
        """Initialize the remote app."""
        super(CesnetOpenIdRemote, self).__init__()

        self.extras_serializer = LocalProxy(lambda: obj_or_import_string(
            current_app.config.get(
                'OAUTHCLIENT_CESNET_OPENID_EXTRA_DATA_SERIALIZER', self.fetch_extra_data)))

    def remote_app(self) -> dict:
        """Configure and return remote app."""
        conf = super(CesnetOpenIdRemote, self).remote_app()
        conf.update(dict(
            response_handler=default_remote_response_handler,
            authorized_redirect_url='/oauth/complete/',
            error_redirect_url='/error',
            disconnect_redirect_url='/logged-out',
            logout_url='https://login.cesnet.cz/oidc/endsession'))
        return conf

    def _account_setup(self, remote, token, refresh=False):
        if refresh:
            session.pop('user_info')

        user_info = self.get_userinfo(remote)
        user_id = self.get_user_id(remote, email=user_info.email)

        with db.session.begin_nested():
            # Update extra_data.
            groups = self.remote_groups_and_extra_data(token.remote_account,
                                                       dict(user_info=user_info, user_id=user_id))
            assert not isinstance(g.identity, AnonymousIdentity)

            user = token.remote_account.user
            current_cesnet_openid.sync_user_roles(user, groups)

        return user_id, user, groups

    def handle_authorized(self, resp, remote, *args, **kwargs):
        """Handle user authorization.
        :param resp: User authorization response
        :param remote: The remote application
        """
        from invenio_oauthclient.handlers.rest import authorized_signup_handler
        response = authorized_signup_handler(resp, remote, *args, **kwargs)
        token = RemoteToken.get_by_token(client_id=remote.consumer_key,
                                         access_token=token_getter(remote)[0])

        if token:
            last_update = token.remote_account.extra_data.get('updated', datetime.utcnow())
            refresh_timedelta = current_app.config['CESNET_OPENID_REMOTE_REFRESH_TIMEDELTA']

            if last_update < (datetime.utcnow() + refresh_timedelta).isoformat():
                self._account_setup(remote, token, refresh=True)
                db.session.commit()
        return response

    def handle_signup(self, remote, *args, **kwargs):
        """Handle signup.
        :param remote: The remote application
        """
        from invenio_oauthclient.handlers.rest import signup_handler
        return signup_handler(remote, *args, **kwargs)

    def account_setup(self, remote, token, resp):
        """Perform additional setup after user have been logged in."""
        token.remote_account.extra_data = {}
        user_id, user, _ = self._account_setup(remote, token)

        # Create user <-> external id link.
        oauth_link_external_id(user, dict(id=user_id, method=self.name))

    def fetch_extra_data(self, user_info, user_id):
        """"Fetch extra account data from resource."""
        extra_data = user_info

        group_uris = extra_data.pop(OPENIDC_GROUPS_SCOPE, [])
        group_uris = [u for u in group_uris if current_cesnet_openid.validate_group_uri(u)]
        extra_data[OPENIDC_GROUPS_KEY] = group_uris

        extra_data['external_id'] = user_id
        extra_data['external_method'] = self.name
        return extra_data

    def remote_groups_and_extra_data(self, account, resource, refresh_timedelta=None):
        """Fetch account roles and extra data from resource if necessary."""
        updated = datetime.utcnow()
        modified_since = updated
        if refresh_timedelta is not None:
            modified_since += refresh_timedelta
        modified_since = modified_since.isoformat()
        last_update = account.extra_data.get("updated", modified_since)

        if last_update > modified_since:
            return account.extra_data.get(OPENIDC_GROUPS_KEY, [])

        user_info = resource.get('user_info')
        user_id = resource.pop('user_id')
        extra_data = self.extras_serializer(user_info, user_id)

        groups = extra_data.get(OPENIDC_GROUPS_KEY, [])
        account.extra_data.update(
            updated=updated.isoformat(), **extra_data
        )
        return groups

    def _disconnect(self, remote, *args, **kwargs):
        """Handle unlinking of remote account.

           :param remote: The remote application.
        """
        if not current_user.is_authenticated:
            return current_app.login_manager.unauthorized()

        account = RemoteAccount.get(user_id=current_user.get_id(),
                                    client_id=remote.consumer_key)
        sub = account.extra_data.get('sub')
        user_info = self.get_userinfo(remote)
        resource = dict(user_info=user_info,
                        user_id=self.get_user_id(remote, email=user_info.email))

        groups = self.remote_groups_and_extra_data(account, resource)
        roles = current_cesnet_openid.groups_roles(groups)
        current_cesnet_openid.remove_user_roles(current_user, roles, reverse=True)

        if sub:
            oauth_unlink_external_id({'id': sub, 'method': self.name})
        if account:
            with db.session.begin_nested():
                account.delete()

    def handle_disconnect(self, remote, *args, **kwargs):
        """Handle unlinking of remote account."""
        self._disconnect(remote, *args, **kwargs)
        redirect_url = current_app.config["OAUTHCLIENT_REST_REMOTE_APPS"][
            remote.name
        ]["disconnect_redirect_url"]
        return response_handler(remote, redirect_url)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CESNET OIDC Auth backend for OARepo"""

OPENIDC_CONFIG = 'CESNET_OPENIDC_CONFIG'
OPENIDC_GROUPS_SCOPE = 'eduperson_entitlement_extended'
OPENIDC_BASE_URL = 'https://login.cesnet.cz/oidc/'
OPENIDC_GROUPS_KEY = 'cesnet_groups'



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Helper proxy to the state object."""

from flask import current_app
from werkzeug.local import LocalProxy

from cesnet_openid_remote.api import CesnetOpenIdRemoteAPI

current_cesnet_openid: CesnetOpenIdRemoteAPI = LocalProxy(
    lambda: current_app.extensions['cesnet-openid-remote'].api
)
"""Proxy to the current CesnetOpenIdRemoteAPI."""



#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Initial cesnet openid remote revision."""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '7b18e6f52083'
down_revision = None
branch_labels = ('cesnet_openid_remote',)
depends_on = None


def upgrade():
    """Upgrade database."""
    pass


def downgrade():
    """Downgrade database."""
    pass



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""
import os
import shutil
import tempfile

import pytest
from flask import Flask
from flask_oauthlib.client import OAuth as FlaskOAuth, OAuthResponse
from invenio_accounts import InvenioAccounts
from invenio_accounts.proxies import current_datastore
from invenio_db import InvenioDB, db
from invenio_oauthclient import InvenioOAuthClient, InvenioOAuthClientREST
from invenio_oauthclient.views.client import rest_blueprint
from invenio_openid_connect import InvenioOpenIDConnect
from invenio_openid_connect.views import blueprint as openid_blueprint
from oarepo_communities import OARepoCommunities
from oarepo_communities.api import OARepoCommunity
from sqlalchemy_utils import database_exists, create_database, drop_database

from cesnet_openid_remote import CesnetOpenIdRemote, CESNETOpenIDRemote
from cesnet_openid_remote.constants import OPENIDC_GROUPS_SCOPE
from cesnet_openid_remote.models import CesnetGroup


@pytest.fixture
def base_app(request):
    """Flask application fixture without OAuthClient initialized."""
    # allow HTTP for keycloak tests, and create the KEYCLOAK_REMOTE_APP
    os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"

    instance_path = tempfile.mkdtemp()
    base_app = Flask('testapp')
    base_app.config.update(
        TESTING=True,
        WTF_CSRF_ENABLED=False,
        LOGIN_DISABLED=False,
        CACHE_TYPE='simple',
        OAUTHCLIENT_REST_REMOTE_APPS=dict(
            cesnet=CesnetOpenIdRemote().remote_app(),
        ),
        CESNET_OPENIDC_CONFIG=dict(
            base_url='https://localhost/tests',
            consumer_key='TEST_OIDC_KEY',
            consumer_secret='TEST_OIDC_SECRET',
            scope=f'openid email profile {OPENIDC_GROUPS_SCOPE} isCesnetEligibleLastSeen'
        ),
        OAUTHCLIENT_STATE_EXPIRES=300,
        # use local memory mailbox
        EMAIL_BACKEND='flask_email.backends.locmem.Mail',
        SQLALCHEMY_DATABASE_URI=os.getenv('SQLALCHEMY_DATABASE_URI',
                                          'sqlite://'),
        SERVER_NAME='localhost',
        DEBUG=False,
        SECRET_KEY='TEST',
        SECURITY_DEPRECATED_PASSWORD_SCHEMES=[],
        SQLALCHEMY_TRACK_MODIFICATIONS=False,
        SECURITY_PASSWORD_HASH='plaintext',
        SECURITY_PASSWORD_SCHEMES=['plaintext'],
        APP_ALLOWED_HOSTS=['localhost'],
        USERPROFILES_EXTEND_SECURITY_FORMS=True,
        SECURITY_SEND_REGISTER_EMAIL=False,
        OAUTHCLIENT_CESNET_OPENID_PROTECTED_ROLES=['admin', 'notextising']
    )
    InvenioDB(base_app)
    InvenioAccounts(base_app)

    with base_app.app_context():
        if str(db.engine.url) != 'sqlite://' and \
            not database_exists(str(db.engine.url)):
            create_database(str(db.engine.url))
        db.create_all()

    def teardown():
        with base_app.app_context():
            db.session.close()
            if str(db.engine.url) != 'sqlite://':
                drop_database(str(db.engine.url))
            shutil.rmtree(instance_path)
            db.engine.dispose()

    request.addfinalizer(teardown)

    base_app.test_request_context().push()
    return base_app


@pytest.fixture
def app(base_app):
    """Flask application fixture."""
    FlaskOAuth(base_app)
    InvenioOAuthClient(base_app)
    InvenioOAuthClientREST(base_app)
    InvenioOpenIDConnect(base_app)
    CESNETOpenIDRemote(base_app)

    # Register blueprint
    base_app.register_blueprint(rest_blueprint)
    base_app.register_blueprint(openid_blueprint)
    return base_app


@pytest.fixture
def communities_app(app):
    """Flask application with communities extension."""
    OARepoCommunities(app)

    return app


@pytest.fixture
def group_uris(cesnet_groups):
    """Group URIs fixture."""
    return {
        'exists': 'urn:geant:cesnet.cz:groupAttributes:f0c14f62-b19c-447e-b044-c3098cebb426?displayName=example#perun.cesnet.cz',
        'new': 'urn:geant:cesnet.cz:groupAttributes:f0c14f62-b19c-447e-b044-c3098c3bb426?displayName=new#perun.cesnet.cz',
        'invalid': 'urn:geant:cesnet.cz:group:f0c14f62-b19c-447e-b044-c3098cebb426#perun.cesnet.cz'
    }


@pytest.fixture
def users_fixture(base_app):
    """Flask app with example data used to test models."""
    with base_app.app_context():
        datastore = base_app.extensions['security'].datastore
        datastore.create_user(
            email='test1@oarepo.org',
            password='tester',
            active=True
        )
        datastore.create_user(
            email='test2@oarepo.org',
            password='tester',
            active=True
        )
        datastore.create_user(
            email='test3@oarepo.org',
            password='tester',
            active=True
        )
        datastore.create_user(
            email='john.doe@example.oarepo.org',
            password='tester',
            active=True
        )
        datastore.commit()


@pytest.fixture()
def example_cesnet(request):
    """CESNET example data."""
    file_path = os.path.join(os.path.dirname(__file__),
                             'data/cesnet_openid_response.json')
    with open(file_path) as response_file:
        json_data = response_file.read()

    from jwt import encode
    token = encode(dict(name="John Doe"), key="1234")

    return OAuthResponse(
        resp=None,
        content=json_data,
        content_type='application/json'
    ), dict(
        access_token=token,
        token_type='bearer',
        expires_in=1199,
        refresh_token='test_refresh_token'
    ), dict(
        user=dict(
            email='john.doe@oarepo.org',
            profile=dict(username='abcd1234@einfra.cesnet.cz', full_name='John Doe'),
        ),
        external_id='abcd1234@einfra.cesnet.cz',
        external_method='CESNET eduID Login',
        active=True,
        eduperson_entitlement_extended=[
            "urn:geant:cesnet.cz:group:f0c14f62-b19c-447e-b044-c3098cebb426#perun.cesnet.cz",
            "urn:geant:cesnet.cz:group:8ece6adb-8677-4482-9aec-5a556c646389#perun.cesnet.cz",
            "urn:geant:cesnet.cz:groupAttributes:f0c14f62-b19c-447e-b044-c3098cebb426?displayName=example#perun.cesnet.cz",
            "urn:geant:cesnet.cz:groupAttributes:8ece6adb-8677-4482-9aec-5a556c646389?displayName=example:subgroup#perun.cesnet.cz",
            "urn:geant:cesnet.cz:groupUuid:f0c14f62-b19c-447e-b044-c3098cebb426#perun.cesnet.cz",
            "urn:geant:cesnet.cz:groupUuid:8ece6adb-8677-4482-9aec-5a556c646389#perun.cesnet.cz",
            "urn:mace:cesnet.cz:other-namespace-group"
        ]
    )


@pytest.fixture()
def community():
    current_datastore.commit()

    community = OARepoCommunity.create(
        {'description': 'Community description'},
        title='Test',
        id_='testing-community')
    db.session.commit()

    return community


@pytest.fixture()
def protected_roles():
    return [current_datastore.create_role(name='admin')]


@pytest.fixture()
def cesnet_groups():
    with db.session.begin_nested():
        cg1 = CesnetGroup(
            uuid='f0c14f62-b19c-447e-b044-c3098cebb426',
            display_name='example',
            uri='urn:geant:cesnet.cz:groupAttributes:f0c14f62-b19c-447e-b044-c3098cebb426?displayName=example#perun.cesnet.cz')
        db.session.add(cg1)

        cg2 = CesnetGroup(
            uuid='8ece6adb-8677-4482-9aec-5a556c646389',
            display_name='example:subgroup',
            uri='urn:geant:cesnet.cz:groupAttributes:8ece6adb-8677-4482-9aec-5a556c646389?displayName=example:subgroup#perun.cesnet.cz')
        db.session.add(cg2)
    return [cg1, cg2]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CESNET OIDC Auth backend for OARepo"""
from functools import wraps

import sqlalchemy
from invenio_accounts.models import Role, User
from invenio_accounts.proxies import current_datastore
from invenio_db import db
from invenio_oauthclient.models import RemoteAccount
from sqlalchemy import func, cast, text

from cesnet_openid_remote.constants import OPENIDC_GROUPS_KEY
from cesnet_openid_remote.errors import OAuthCESNETInvalidGroupURI, OAuthCESNETGroupExists, OAuthCESNETRoleProtected
from cesnet_openid_remote.groups import validate_group_uri, parse_group_uri
from cesnet_openid_remote.models import CesnetGroup


def check_role_protected(f):
    """Decorator checking if passed role is protected."""

    @wraps(f)
    def inner(self, *args, **kwargs):
        for arg in args:
            if isinstance(arg, Role) and arg in self.protected_roles:
                raise OAuthCESNETRoleProtected

        return f(self, *args, **kwargs)

    return inner


class CesnetOpenIdRemoteAPI(object):
    """Cesnet OIDC Remote API."""

    def __init__(self, app, remote):
        self.app = app
        self.remote_app = remote
        self.validate_group_uri = app.config.get(
            'OAUTHCLIENT_CESNET_OPENID_GROUP_VALIDATOR', validate_group_uri)
        self.parse_group_uri = app.config.get(
            'OAUTHCLIENT_CESNET_OPENID_GROUP_PARSER', parse_group_uri)

    @property
    def protected_roles(self):
        protected_role_names = self.app.config.get(
            'OAUTHCLIENT_CESNET_OPENID_PROTECTED_ROLES', [])
        protected_roles = [current_datastore.find_role(r) for r in protected_role_names]
        return list(filter(None, protected_roles))

    @classmethod
    def find_group(cls, uuid: str) -> CesnetGroup:
        """Find a CESNET group by its unique identifier.

            :param uuid: The unique external identifier of the group
            :returns CesnetGroup: group instance, None if not found
        """
        return CesnetGroup.query.filter_by(uuid=uuid).first()

    def find_or_create_group(self, uri: str) -> CesnetGroup:
        """Find a CESNET group by its URI, or create a new one.

            :param uri: External group URI
            :returns CesnetGroup: found or created CesnetGroup object
        """
        if not validate_group_uri(uri):
            raise OAuthCESNETInvalidGroupURI(uri)

        guuid, attrs = self.parse_group_uri(uri)
        return self.find_group(guuid) or self.create_group(uuid=guuid, attrs=attrs, uri=uri)

    @classmethod
    def remote_accounts(cls, group: CesnetGroup) -> list:
        """Find all remote accounts that are members of a given cesnet group.

            :param group: CesnetGroup to list all members of
            :returns list of all group members
        """
        if db.session.bind.dialect.name != 'postgresql':
            group_field = cast(RemoteAccount.extra_data, sqlalchemy.String)
            query = RemoteAccount.query.filter(group_field.contains(group.uri))
        else:
            query = RemoteAccount.query.filter(
                text(f"(extra_data->'{OPENIDC_GROUPS_KEY}')::jsonb ? '{group.uri}'"))

        return query.all()

    @classmethod
    def remote_group_uris(cls, account: RemoteAccount):
        """Returns list of all CESNET group URIs for a given RemoteAccount."""
        return account.extra_data[OPENIDC_GROUPS_KEY]

    @classmethod
    def create_group(cls, uuid: str, attrs: dict, uri: str) -> CesnetGroup:
        """Create a CESNET group record from external group URI.

           :param uuid: External group unique UUID identifier
           :param attrs: External group additional attributes
           :param uri: External group URI
           :returns CesnetGroup: created cesnet group object
           :raises OAuthCESNETInvalidGroupURI:
        """
        display_name = attrs.get('displayName', None)
        with db.session.begin_nested():
            if CesnetGroup.query.filter_by(uuid=uuid).count():
                raise OAuthCESNETGroupExists(uuid)

            cg = CesnetGroup(uuid=uuid,
                             display_name=display_name,
                             uri=uri)
            db.session.add(cg)
            return cg

    def groups_roles(self, group_uris: list) -> list:
        """Return a list of all Invenio Roles associated with group URIs.

            :param group_uris: List of CESNET group URIs
            :returns list[Role]: list of all associated Invenio roles
        """
        group_roles = []
        for uri in group_uris:
            group_roles += self.find_or_create_group(uri).roles

        return group_roles

    @check_role_protected
    def add_group_role(self, group: CesnetGroup, role: Role):
        """Assign a CESNET group to Invenio role.

            :param group: CESNET group to be assigned
            :param role: target Invenio role to be assigned to group
            :returns True if group was assigned, False otherwise
        """
        if role not in group.roles:
            with db.session.begin_nested():
                group.roles.append(role)

                # Find and add to role all remote accounts with this group
                for ra in self.remote_accounts(group):
                    self.sync_user_roles(ra.user, self.remote_group_uris(ra))

    @check_role_protected
    def remove_group_role(self, group: CesnetGroup, role: Role):
        """Remove CESNET group from Invenio role.

            :param group: CESNET group to be removed
            :param role: Invenio role
            :raises OAuthCESNETGroupNotInRole:
        """
        if role in group.roles:
            with db.session.begin_nested():
                group.roles.remove(role)

                # Find and remove from role all remote account users
                # with this group that haven't any other group assigned to this role
                for ra in self.remote_accounts(group):
                    self.sync_user_roles(ra.user, self.remote_group_uris(ra))

    def add_user_roles(self, user: User, group_roles: list):
        """Assign Invenio roles to user based on his cesnet groups.

            :param user: User instance to be roles assigned to
            :param group_roles: list of user group's roles
        """
        group_roles = set(group_roles) - set(self.protected_roles)
        with db.session.begin_nested():
            for role in group_roles:
                current_datastore.add_role_to_user(user, role)

    def remove_user_roles(self, user: User, group_roles: list, reverse=False):
        """Remove user from any role that is not in group-granted roles.

            :param user: User instance to remove roles from
            :param group_roles: list of user group's roles
            :param reverse: if True, removes roles granted by group roles instead
        """
        if not reverse:
            extra_roles = (set(user.roles) - set(group_roles)) - set(self.protected_roles)
        else:
            extra_roles = set(user.roles).union(set(group_roles)) - set(self.protected_roles)
        with db.session.begin_nested():
            for extra in extra_roles:
                current_datastore.remove_role_from_user(user, extra)

    def sync_user_roles(self, user: User, group_uris: list):
        """Synchronize user's roles based on his cesnet group URIs."""

        roles = self.groups_roles(group_uris)

        self.remove_user_roles(user, roles)
        self.add_user_roles(user, roles)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CESNET OIDC Auth backend for OARepo"""
from invenio_oauthclient.errors import OAuthResponseError


class OAuthCESNETInvalidGroupURI(Exception):
    """External group URI did not match the expected pattern."""

    def __init__(self, uri, *args, **kwargs):
        super(OAuthCESNETInvalidGroupURI, self).__init__('Invalid external group URI',
                                                         uri, *args)


class OAuthCESNETRoleProtected(Exception):
    """External group URI did not match the expected pattern."""


class OAuthCESNETGroupExists(Exception):
    """External group with this URI already exists."""

    def __init__(self, uuid, *args, **kwargs):
        super(OAuthCESNETGroupExists, self).__init__('External group with this UUID already exists',
                                                     uuid, *args)


class OAuthCESNETRejectedAccountError(OAuthResponseError):
    """Define exception for not allowed cesnet group accounts."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Module tests."""
import pytest
from invenio_accounts.models import Role, User
from invenio_accounts.proxies import current_datastore
from invenio_db import db

from cesnet_openid_remote.api import CesnetOpenIdRemoteAPI
from cesnet_openid_remote.constants import OPENIDC_GROUPS_KEY
from cesnet_openid_remote.errors import OAuthCESNETInvalidGroupURI, OAuthCESNETGroupExists, OAuthCESNETRoleProtected
from cesnet_openid_remote.models import CesnetGroup
from cesnet_openid_remote.proxies import current_cesnet_openid
from tests.test_remote import _login_user


def test_init(app, protected_roles):
    api = CesnetOpenIdRemoteAPI(app, None)
    assert len(api.protected_roles) == len(protected_roles)
    assert api.protected_roles[0] == protected_roles[0]


def test_find_group(app, group_uris):
    assert len(CesnetGroup.query.all()) == 2

    # Test find existing group
    gr = current_cesnet_openid.find_group('f0c14f62-b19c-447e-b044-c3098cebb426')
    assert gr is not None
    assert str(gr.uuid) == 'f0c14f62-b19c-447e-b044-c3098cebb426'
    assert gr.display_name == 'example'

    # Test find non-existing group
    gr = current_cesnet_openid.find_group('e0c14f62-b19c-447e-b044-c3098cebb42e')
    assert gr is None


def test_find_or_create_group(app, group_uris):
    # Test find existing group
    gr = current_cesnet_openid.find_or_create_group(group_uris['exists'])
    assert gr is not None
    assert str(gr.uuid) == 'f0c14f62-b19c-447e-b044-c3098cebb426'
    assert gr.display_name == 'example'

    # Test create group
    gr = current_cesnet_openid.find_or_create_group(group_uris['new'])
    assert gr is not None
    assert str(gr.uuid == 'f0c14f62-b19c-447e-b044-c3098c3bb426')
    assert gr.display_name == 'new'

    # Test find invalid URI
    with pytest.raises(OAuthCESNETInvalidGroupURI):
        current_cesnet_openid.find_or_create_group(group_uris['invalid'])


def test_group_create(app, group_uris):
    """Test cesnet group creation using api."""

    # Test create from valid uri
    assert len(CesnetGroup.query.all()) == 2
    res = current_cesnet_openid.create_group(
        'f0c14f62-b19c-447e-b044-c3098ce2b426',
        {'displayName': 'example'}, group_uris['new'])

    assert isinstance(res, CesnetGroup)
    groups = CesnetGroup.query.all()
    assert len(groups) == 3
    assert groups[2] == res
    assert res.display_name == 'example'
    assert res.uri == group_uris['new']
    assert res.uuid == 'f0c14f62-b19c-447e-b044-c3098ce2b426'
    assert len(res.roles) == 0

    # Test create already created group
    with pytest.raises(OAuthCESNETGroupExists):
        current_cesnet_openid.create_group(
            'f0c14f62-b19c-447e-b044-c3098cebb426',
            {'displayName': 'nope'}, group_uris['exists'])


def test_add_group_to_role(communities_app, cesnet_groups, example_cesnet, community, protected_roles, users_fixture):
    """Test adding a cesnet group to invenio role."""
    roles = Role.query.all()
    admin = current_datastore.find_role('admin')
    assert len(roles) == 4
    assert len(CesnetGroup.query.all()) == 2

    group = CesnetGroup.query.first()
    assert len(group.roles) == 0

    # Test add not assigned cesnet group
    current_cesnet_openid.add_group_role(group, roles[0])
    assert len(group.roles) == 1
    assert len(roles[0].cesnet_groups.all()) == 1
    assert group.roles[0] == roles[0]

    # Test add already added group
    current_cesnet_openid.add_group_role(group, roles[0])
    assert len(group.roles) == 1
    assert group.roles[0] == roles[0]

    # Test add to protected roles fails miserably
    with pytest.raises(OAuthCESNETRoleProtected):
        current_cesnet_openid.add_group_role(group, admin)

    # Test remote account role is added on group role addition
    user = User.query.filter_by(email='john.doe@example.oarepo.org').first()
    with communities_app.test_client() as c:
        _login_user(communities_app, example_cesnet, c)
        ra = current_cesnet_openid.remote_accounts(group)[0]
        assert len(ra.user.roles) == 1
        assert ra.user.roles[0] == roles[0]

        current_cesnet_openid.add_group_role(group, roles[1])
        assert len(group.roles) == 2
        assert len(ra.user.roles) == 2
        assert roles[1] in ra.user.roles


def test_rm_role_from_group(communities_app, cesnet_groups, community, protected_roles, example_cesnet, users_fixture):
    """Test removing a role from a cesnet group."""
    group = CesnetGroup.query.first()
    roles = Role.query.all()
    admin = current_datastore.find_role('admin')
    current_cesnet_openid.add_group_role(group, roles[0])
    assert len(group.roles) == 1

    # Remove group from assigned roles
    current_cesnet_openid.remove_group_role(group, roles[0])
    assert len(group.roles) == 0

    # Remove group from non assigned role
    current_cesnet_openid.remove_group_role(group, roles[0])
    assert len(group.roles) == 0

    # Test remove from protected role fails miserably
    group.roles.append(admin)
    assert admin in group.roles
    with pytest.raises(OAuthCESNETRoleProtected):
        current_cesnet_openid.remove_group_role(group, admin)

    # Test remote account role is removed on group role removal
    user = User.query.filter_by(email='john.doe@example.oarepo.org').first()
    group.roles.append(roles[0])
    user.roles.append(roles[0])

    with communities_app.test_client() as c:
        _login_user(communities_app, example_cesnet, c)
        ra = current_cesnet_openid.remote_accounts(group)[0]

        assert len(ra.user.roles) == 1
        assert ra.user.roles[0] == roles[0]
        current_cesnet_openid.remove_group_role(group, roles[0])
        assert set(group.roles) == {admin}
        assert len(ra.user.roles) == 0


def test_add_user_roles(communities_app, community, cesnet_groups, protected_roles, users_fixture):
    """Test role assignment based on user's cesnet groups."""
    group = CesnetGroup.query.first()
    role = Role.query.first()
    user = User.query.filter_by(email='john.doe@example.oarepo.org').first()
    group.roles.append(role)

    assert len(user.roles) == 0

    # Test add to valid role granted by group
    current_cesnet_openid.add_user_roles(user, group.roles)
    assert len(user.roles) == 1
    assert user.roles[0] == role

    # Test add roles from group doesn't add user to protected_role
    group.roles.append(protected_roles[0])
    current_cesnet_openid.add_user_roles(user, group.roles)
    assert len(user.roles) == 1
    assert protected_roles[0] not in user.roles


def test_remove_user_roles(communities_app, community, cesnet_groups, protected_roles, users_fixture):
    """Test removal from roles not granted by any of user's groups."""
    group = CesnetGroup.query.first()
    role = Role.query.first()
    extra_role = Role.query.all()[1]
    user = User.query.filter_by(email='john.doe@example.oarepo.org').first()

    group.roles.append(role)
    user.roles.append(role)
    user.roles.append(extra_role)
    assert len(user.roles) == 2

    # Test remove regular extra roles from user
    current_cesnet_openid.remove_user_roles(user, group.roles)
    assert len(user.roles) == 1
    assert extra_role not in user.roles

    # Test that protected role is not stripped from user
    user.roles.append(protected_roles[0])
    current_cesnet_openid.remove_user_roles(user, group.roles)
    assert len(user.roles) == 2
    assert protected_roles[0] in user.roles

    # Test reversed role removal
    user.roles.append(extra_role)
    assert role in user.roles
    assert role in group.roles

    current_cesnet_openid.remove_user_roles(user, group.roles, reverse=True)
    assert role in group.roles
    assert role not in user.roles
    assert extra_role not in user.roles
    assert len(user.roles) == 1


def test_sync_user_roles(communities_app, community, cesnet_groups, protected_roles, users_fixture):
    """Test synchronization of user roles based on his cesnet groups."""
    group = CesnetGroup.query.first()
    role = Role.query.first()
    extra_role = Role.query.all()[1]
    user = User.query.filter_by(email='john.doe@example.oarepo.org').first()

    group.roles.append(role)
    user.roles.append(extra_role)

    # Test sync adds group role and removes extra role
    current_cesnet_openid.sync_user_roles(user, [group.uri])
    assert len(user.roles) == 1
    assert extra_role not in user.roles
    assert role in user.roles


def test_remote_accounts(communities_app, community, example_cesnet, cesnet_groups, users_fixture):
    group = CesnetGroup.query.first()
    role = Role.query.first()
    user = User.query.filter_by(email='john.doe@example.oarepo.org').first()

    group.roles.append(role)
    user.roles.append(role)

    with communities_app.test_client() as c:
        _login_user(communities_app, example_cesnet, c)

        ras = current_cesnet_openid.remote_accounts(group)
        assert len(ras) == 1
        assert ras[0].user == user
        assert group.uri in ras[0].extra_data[OPENIDC_GROUPS_KEY]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CESNET OIDC Auth backend for OARepo"""
from flask_babelex import gettext
from invenio_db import db
from speaklater import make_lazy_gettext
from sqlalchemy_utils import UUIDType

_ = make_lazy_gettext(lambda: gettext)


cesnet_group_role = db.Table(
    'cesnet_group_roles',
    db.Column('group_id', db.Integer(), db.ForeignKey(
        'cesnet_group.id', name='fk_cesnet_group_roles_group_id')),
    db.Column('role_id', db.Integer(), db.ForeignKey(
        'accounts_role.id', name='fk_cesnet_group_roles_role_id')),
)
"""Relationship between CESNET groups and Invenio roles."""


class CesnetGroup(db.Model):
    """Cesnet external group data model."""

    __tablename__ = 'cesnet_group'

    id = db.Column(db.Integer(), primary_key=True)
    uuid = db.Column(UUIDType(), nullable=False, unique=True, index=True)
    display_name = db.Column(db.String(255))
    uri = db.Column(db.String(1024), unique=True, nullable=False)
    roles = db.relationship('Role', secondary=cesnet_group_role,
                            backref=db.backref('cesnet_groups', lazy='dynamic'))

    def __str__(self):
        """Return the name and description of the role."""
        return 'CesnetGroup<{0.uuid}> {0.display_name}'.format(self)

    def __repr__(self):
        return self.__str__()





# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CESNET OIDC Auth backend for OARepo"""

from .remote import CesnetOpenIdRemote
from .ext import CESNETOpenIDRemote
from .version import __version__

__all__ = ('__version__', 'CesnetOpenIdRemote', 'CESNETOpenIDRemote')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Version information for CESNET-OpenID-Remote.

This file is imported by ``cesnet_openid_remote.__init__``,
and parsed by ``setup.py``.
"""

__version__ = '2.0.1'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CESNET OIDC Auth backend for OARepo"""
from datetime import timedelta

from flask import current_app, session, g
from flask_login import current_user, user_logged_out
from flask_principal import identity_loaded, AnonymousIdentity, identity_changed, RoleNeed, UserNeed
from invenio_oauthclient.models import RemoteAccount
from invenio_oauthclient.utils import obj_or_import_string
from werkzeug.local import LocalProxy

from cesnet_openid_remote.constants import OPENIDC_GROUPS_KEY
from cesnet_openid_remote.proxies import current_cesnet_openid

CESNET_OPENID_REMOTE_SESSION_KEY = 'identity.cesnet_provides'
"""Name of session key where CESNET roles are stored."""

CESNET_OPENID_REMOTE_REFRESH_TIMEDELTA = timedelta(minutes=-5)
"""Default interval for refreshing user's extra data (e.g. groups)."""

sconf = LocalProxy(
    lambda: dict(key=current_app.config.get(
        "CESNET_OPENID_REMOTE_SESSION_KEY",
        CESNET_OPENID_REMOTE_SESSION_KEY),
        refresh=current_app.config.get(
            "CESNET_OPENID_REMOTE_REFRESH_TIMEDELTA",
            CESNET_OPENID_REMOTE_REFRESH_TIMEDELTA)))


@identity_changed.connect
def on_identity_changed(sender, identity):
    """Store roles in session whenever identity changes.

    :param sender: Sender of the signal
    :param identity: The user identity where information are stored.
    """
    remote = current_cesnet_openid.remote_app

    if isinstance(identity, AnonymousIdentity):
        return

    logged_in_via_token = \
        hasattr(current_user, 'login_via_oauth2') \
        and getattr(current_user, 'login_via_oauth2')

    client_id = remote.get_consumer_key()
    remote_account = RemoteAccount.get(
        user_id=identity.id, client_id=client_id
    )

    groups = []
    if remote_account and not logged_in_via_token:
        if sconf['refresh']:
            user_info = remote.get_userinfo(remote)
            resource = dict(user_info=user_info,
                            user_id=remote.get_user_id(remote, email=user_info.email))
            groups = remote.remote_groups_and_extra_data(
                remote_account, resource, refresh_timedelta=sconf['refresh']
            )
        else:
            groups = remote_account.extra_data[OPENIDC_GROUPS_KEY]
    elif remote_account and logged_in_via_token:
        groups = remote_account.extra_data[OPENIDC_GROUPS_KEY]

    current_cesnet_openid.sync_user_roles(current_user, groups)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CESNET OIDC Auth backend for OARepo"""
from invenio_oauthclient.utils import obj_or_import_string
from invenio_openid_connect.signals import prepare_state_view_data

from . import config
from .api import CesnetOpenIdRemoteAPI
from .constants import OPENIDC_CONFIG
from .state import transform_state_data
from .remote import InvenioAuthOpenIdRemote


class CESNETOpenIDRemote(object):
    """CESNET-OpenID-Remote extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        self.connect_signals(app)

        remote = InvenioAuthOpenIdRemote()
        app.extensions['cesnet-openid-remote'] = self
        app.extensions['cesnet-openid-remote'].api = CesnetOpenIdRemoteAPI(app, remote)

    def init_config(self, app):
        """Initialize configuration."""
        for k in dir(config):
            if k.startswith('CESNET_OPENID_REMOTE_') or \
                k in ['OAUTHCLIENT_REST_REMOTE_APPS', OPENIDC_CONFIG]:
                app.config.setdefault(k, getattr(config, k))

    def connect_signals(self, app):
        state_transformer = obj_or_import_string(
            app.config.get(
                "OAUTHCLIENT_CESNET_OPENID_STATE_TRANSFORM", transform_state_data))
        prepare_state_view_data.connect(state_transformer)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Module tests."""

from flask import Flask
from flask_oauthlib.client import OAuth as FlaskOAuth
from invenio_oauthclient import InvenioOAuthClient
from invenio_openid_connect.remote import LazyOAuthRemoteApp

from cesnet_openid_remote import CESNETOpenIDRemote, CesnetOpenIdRemote


def test_version():
    """Test version import."""
    from cesnet_openid_remote import __version__
    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask('testapp')
    ext = CESNETOpenIDRemote(app)
    assert 'cesnet-openid-remote' in app.extensions

    app = Flask('testapp')
    ext = CESNETOpenIDRemote()
    assert 'cesnet-openid-remote' not in app.extensions
    ext.init_app(app)
    assert 'cesnet-openid-remote' in app.extensions


class _CustomCesnetRemoteApp(CesnetOpenIdRemote):
    """Custom OAuthRemoteApp used for testing."""


def test_standard_remote_app_factory(base_app):
    """Test standard remote_app class."""
    base_app.config.update(
        OAUTHCLIENT_REMOTE_APPS=dict(
            custom_app=_CustomCesnetRemoteApp().remote_app()
        )
    )
    FlaskOAuth(base_app)
    InvenioOAuthClient(base_app)
    CESNETOpenIDRemote(base_app)
    assert isinstance(
        base_app.extensions['oauthlib.client'].remote_apps['custom_app'],
        LazyOAuthRemoteApp)
    assert not isinstance(
        base_app.extensions['oauthlib.client'].remote_apps['custom_app'],
        _CustomCesnetRemoteApp)



#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create cesnet openid remote tables."""
import sqlalchemy_utils
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0f27e899fc55'
down_revision = '7b18e6f52083'
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('cesnet_group',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('uuid', sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
    sa.Column('display_name', sa.String(length=255), nullable=False),
    sa.Column('uri', sa.String(length=1024), nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_cesnet_group')),
    sa.UniqueConstraint('uri', name=op.f('uq_cesnet_group_uri'))
    )
    op.create_index(op.f('ix_cesnet_group_uuid'), 'cesnet_group', ['uuid'], unique=True)
    op.create_table('cesnet_group_roles',
    sa.Column('group_id', sa.Integer(), nullable=True),
    sa.Column('role_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['group_id'], ['cesnet_group.id'], name='fk_cesnet_group_roles_group_id'),
    sa.ForeignKeyConstraint(['role_id'], ['accounts_role.id'], name='fk_cesnet_group_roles_role_id')
    )
    # ### end Alembic commands ###


def downgrade():
    """Downgrade database."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('cesnet_group_roles')
    op.drop_index(op.f('ix_cesnet_group_uuid'), table_name='cesnet_group')
    op.drop_table('cesnet_group')
    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CESNET OIDC Auth backend for OARepo"""
import click
from click import UUID
from flask.cli import with_appcontext
from invenio_accounts.proxies import current_datastore
from invenio_db import db
from oarepo_micro_api.cli import with_api

from cesnet_openid_remote.models import CesnetGroup
from cesnet_openid_remote.proxies import current_cesnet_openid


def _prepare_group_role(group_uuid, role_name):
    group = current_cesnet_openid.find_group(group_uuid)
    if not group:
        click.secho(f'CESNET group with UUID {group_uuid} does not exist', fg='red')
        exit(2)
    role = current_datastore.find_role(role_name)
    if not role:
        click.secho(f'Role {role_name} does not exist', fg='red')
        exit(3)
    return group, role


@click.group()
def cesnet_group():
    """Management commands for CESNET external group mappings."""


@cesnet_group.command('create')
@click.argument('uri')
@with_appcontext
@with_api
def create(uri):
    """Create an external CESNET group."""
    if not current_cesnet_openid.validate_group_uri(uri):
        click.secho(f'Group URI is invalid', fg='red')

    uuid, attrs = current_cesnet_openid.parse_group_uri(uri)
    created = current_cesnet_openid.create_group(uuid, attrs, uri)
    db.session.commit()

    click.secho(f'Successfully Created: {created}', fg='green')


@cesnet_group.command('add')
@click.argument('group_uuid')
@click.argument('role_name')
@with_appcontext
@with_api
def add(group_uuid, role_name):
    """Add a CESNET group to Invenio Role."""
    group, role = _prepare_group_role(group_uuid, role_name)
    current_cesnet_openid.add_group_role(group, role)
    db.session.commit()
    click.secho(f'Roles assigned with {group}: {group.roles}', fg='green')


@cesnet_group.command('remove')
@click.argument('group_uuid')
@click.argument('role_name')
@with_appcontext
@with_api
def remove(group_uuid, role_name):
    """Remove a CESNET group from an Invenio Role."""
    group, role = _prepare_group_role(group_uuid, role_name)
    current_cesnet_openid.remove_group_role(group, role)
    db.session.commit()
    roles = [r.name for r in group.roles]
    click.secho(f'Roles assigned with {group}: {roles}', fg='green')


@cesnet_group.command('list')
@with_appcontext
@with_api
def list_groups():
    """List external CESNET groups."""
    groups = CesnetGroup.query.all()
    for g in groups:
        roles = [r.name for r in g.roles]
        click.secho('{0.uuid} {0.display_name} -> {1}'.format(g, roles))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CESNET OIDC Auth backend for OARepo"""

import os

from setuptools import find_packages, setup

readme = open('README.md').read()
history = open('CHANGES.rst').read()

OAREPO_VERSION = os.environ.get('OAREPO_VERSION', '3.3.0')

tests_require = [
    'pydocstyle',
    'isort',
    'oarepo-communities>=1.1.0',
    'invenio-oauthclient==1.4.0'
]

extras_require = {
    'tests': [
        'oarepo[tests]~={version}'.format(version=OAREPO_VERSION),
        *tests_require
    ]
}

extras_require['all'] = []
for reqs in extras_require.values():
    extras_require['all'].extend(reqs)

setup_requires = [
]

install_requires = [
    'invenio-openid-connect>=2.1.0',
]

packages = find_packages()


# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join('cesnet_openid_remote', 'version.py'), 'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

setup(
    name='cesnet-openid-remote',
    version=version,
    description=__doc__,
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    keywords='invenio oarepo oauth openidc auth groups',
    license='MIT',
    author='Miroslav Bauer',
    author_email='bauer@cesnet.cz',
    url='https://github.com/oarepo/cesnet-openid-remote',
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    entry_points={
        'flask.commands': [
            'cesnet:group = cesnet_openid_remote.cli:cesnet_group',
        ],
        'invenio_base.apps': [
            'cesnet_openid_remote = cesnet_openid_remote:CESNETOpenIDRemote',
        ],
        # TODO: Edit these entry points to fit your needs.
        # 'invenio_access.actions': [],
        # 'invenio_admin.actions': [],
        # 'invenio_assets.bundles': [],
        'invenio_base.api_apps': [
            'cesnet_openid_remote = cesnet_openid_remote:CESNETOpenIDRemote',
        ],
        # 'invenio_base.api_blueprints': [],
        # 'invenio_base.blueprints': [],
        # 'invenio_celery.tasks': [],
        'invenio_db.models': [
            'cesnet_openid_remote = cesnet_openid_remote.models',
        ],
        'invenio_db.alembic': [
            'cesnet_openid_remote = cesnet_openid_remote:alembic',
        ],
        # 'invenio_pidstore.minters': [],
        # 'invenio_records.jsonresolver': [],
    },
    extras_require=extras_require,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Development Status :: 1 - Planning',
    ],
)



#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Make group display name optional."""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b2f210a5a578'
down_revision = '0f27e899fc55'
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('cesnet_group', 'display_name',
               existing_type=sa.VARCHAR(length=255),
               nullable=True)
    # ### end Alembic commands ###


def downgrade():
    """Downgrade database."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('cesnet_group', 'display_name',
               existing_type=sa.VARCHAR(length=255),
               nullable=False)
    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Module tests."""
from flask import url_for, g, session
from flask_security import login_user, logout_user
from invenio_accounts.models import Role
from invenio_accounts.proxies import current_datastore
from invenio_oauthclient.models import RemoteAccount
from invenio_openid_connect.utils import get_dict_from_response

from cesnet_openid_remote import CesnetOpenIdRemote
from cesnet_openid_remote.config import CESNET_OPENID_REMOTE_SESSION_KEY
from cesnet_openid_remote.constants import OPENIDC_GROUPS_KEY
from cesnet_openid_remote.models import CesnetGroup
from tests.helpers import mock_response, mock_remote_get, get_state


def _login_user(app, example_cesnet, c):
    ioc = app.extensions['oauthlib.client']

    # Ensure remote apps have been loaded (due to before first request)
    resp = c.get(url_for('invenio_oauthclient.rest_login',
                         remote_app='cesnet'))
    assert resp.status_code == 302

    example_response, example_token, example_account_info = \
        example_cesnet

    mock_response(app.extensions['oauthlib.client'], 'cesnet',
                  example_token)
    mock_remote_get(ioc, 'cesnet', example_response)

    resp = c.get(url_for(
        'invenio_oauthclient.rest_authorized',
        remote_app='cesnet', code='test',
        state=get_state('cesnet')))
    assert resp.status_code == 302
    assert resp.location == 'http://localhost/oauth/complete/?message=Successfully+authorized.&code=200'


def test_fetch_extra_data(app, example_cesnet, cesnet_groups, users_fixture):
    """Test extra data extraction."""
    example_response, _, _ = example_cesnet
    res = get_dict_from_response(example_response)

    remote = CesnetOpenIdRemote()

    user_info = remote.userinfo_cls(res)
    user_info.username = remote.get_username(user_info)

    extra_data = remote.fetch_extra_data(user_info, user_info['sub'])

    # Test only valid group URIs remained in extra_data
    assert len(extra_data[OPENIDC_GROUPS_KEY]) == 2
    assert set(extra_data[OPENIDC_GROUPS_KEY]) == {
        'urn:geant:cesnet.cz:groupAttributes:f0c14f62-b19c-447e-b044-c3098cebb426?displayName=example#perun.cesnet.cz',
        'urn:geant:cesnet.cz:groupAttributes:8ece6adb-8677-4482-9aec-5a556c646389?displayName=example:subgroup#perun.cesnet.cz'
    }

    assert extra_data['external_id'] == 'abcd1234@einfra.cesnet.cz'


def test_remote_groups_and_extra_data(app, example_cesnet, users_fixture):
    """Test get remote groups and extra data."""
    remote = CesnetOpenIdRemote()
    user = current_datastore.find_user(email='john.doe@example.oarepo.org')

    with app.test_client() as c:
        _login_user(app, example_cesnet, c)

        ra = RemoteAccount.query.first()
        assert ra

        groups = remote.remote_groups_and_extra_data(
            ra, dict(user_info=example_cesnet[2], user_id=user.id))

        assert groups
        assert len(groups) == 2
        assert set(groups) == {
            'urn:geant:cesnet.cz:groupAttributes:f0c14f62-b19c-447e-b044-c3098cebb426?displayName=example#perun.cesnet.cz',
            'urn:geant:cesnet.cz:groupAttributes:8ece6adb-8677-4482-9aec-5a556c646389?displayName=example:subgroup#perun.cesnet.cz'
        }
        assert OPENIDC_GROUPS_KEY in ra.extra_data
        assert ra.extra_data
        assert len(ra.extra_data[OPENIDC_GROUPS_KEY]) == len(groups)


def test_account_setup(communities_app, community, example_cesnet, cesnet_groups, users_fixture):
    """Test account setup after login."""
    group = CesnetGroup.query.first()
    role = Role.query.first()
    group.roles.append(role)

    with communities_app.test_client() as c:
        _login_user(communities_app, example_cesnet, c)

        user = current_datastore.find_user(email='john.doe@example.oarepo.org')
        assert user
        assert len(user.roles) == 1
        assert user.roles[0] == role


def test_disconnect(communities_app, community, cesnet_groups, users_fixture, example_cesnet, protected_roles):
    user = current_datastore.find_user(email='john.doe@example.oarepo.org')
    role = Role.query.first()
    extra_role = Role.query.all()[1]
    group = CesnetGroup.query.first()

    user.roles.append(protected_roles[0])
    user.roles.append(extra_role)
    user.roles.append(role)
    group.roles.append(role)

    assert len(user.roles) == 3
    assert len(group.roles) == 1

    with communities_app.test_client() as c:
        _login_user(communities_app, example_cesnet, c)

        remote = communities_app.extensions['oauthlib.client'].remote_apps['cesnet']
        # Test remove remote account and from remote groups
        CesnetOpenIdRemote().handle_disconnect(remote)
        assert len(user.roles) == 1  # only protected role should remain
        assert protected_roles[0] in user.roles
        assert len(group.roles) == 1
        assert group.roles[0] == role



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CESNET OIDC Auth backend for OARepo"""
from flask import current_app
from invenio_accounts.proxies import current_datastore


def transform_state_data(user, state):
    """Transforms auth state data where necessary."""
    return state



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Module tests."""
from unittest.mock import MagicMock

from invenio_oauthclient._compat import _create_identifier
from invenio_oauthclient.views.client import serializer


def get_state(app='test'):
    """Get state."""
    return serializer.dumps({'app': app, 'sid': _create_identifier(),
                             'next': None, })


def mock_response(oauth, remote_app='test', data=None):
    """Mock the oauth response to use the remote."""
    oauth.remote_apps[remote_app].handle_oauth2_response = MagicMock(
        return_value=data
    )


def mock_remote_get(oauth, remote_app='test', data=None):
    """Mock the oauth remote get response."""
    oauth.remote_apps[remote_app].get = MagicMock(
        return_value=data
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CESNET.
#
# CESNET-OpenID-Remote is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""CESNET OIDC Auth backend for OARepo"""
import os
from datetime import timedelta

from cesnet_openid_remote import CesnetOpenIdRemote
from cesnet_openid_remote.constants import OPENIDC_GROUPS_SCOPE, OPENIDC_BASE_URL

CESNET_OPENIDC_CONFIG = dict(
    base_url=OPENIDC_BASE_URL,
    consumer_key=os.environ.get('OPENIDC_KEY', 'MISSING_OIDC_KEY'),
    consumer_secret=os.environ.get('OPENIDC_SECRET', 'MISSING_OIDC_SECRET'),
    scope=f'openid email profile {OPENIDC_GROUPS_SCOPE} isCesnetEligibleLastSeen'
)

CESNET_OPENID_REMOTE_REFRESH_TIMEDELTA = timedelta(minutes=-5)
"""Default interval for refreshing user's extra data (e.g. groups)."""

CESNET_OPENID_REMOTE_GROUP_PREFIX = 'urn:geant:cesnet.cz:'
"""Default prefix of group attribute URIs."""

CESNET_OPENID_REMOTE_GROUP_AUTHORITY = 'perun.cesnet.cz'
"""Default authority that issues the group attribute URIs."""

CESNET_OPENID_REMOTE_SESSION_KEY = 'identity.cesnet_provides'
"""Name of session key where CESNET roles are stored."""

OAUTHCLIENT_CESNET_OPENID_GROUP_VALIDATOR = 'cesnet_openid_remote.groups.validate_group_uri'
"""Function used to validate external group URIs."""

OAUTHCLIENT_CESNET_OPENID_GROUP_PARSER = 'cesnet_openid_remote.groups.parse_group_uri'
"""Function used to parse external group URIs to (UUID, extra_data) pairs."""

OAUTHCLIENT_CESNET_OPENID_STATE_TRANSFORM = 'cesnet_openid_remote.state.transform_state_data'
"""Function used to validate external group URIs."""

OAUTHCLIENT_CESNET_OPENID_PROTECTED_ROLES = ['admin']
"""Role names that shouldn't be managed/(un)assigned to users by this extension."""

OAUTHCLIENT_REST_REMOTE_APPS = dict(
    eduid=CesnetOpenIdRemote().remote_app(),
)
